# file generated by setuptools_scm
# don't change, don't track in version control
__version__ = version = '0.1.dev79+g41fba5e.d20230612'
__version_tuple__ = version_tuple = (0, 1, 'dev79', 'g41fba5e.d20230612')
